#include <stdlib.h>
#include <sys/types.h>
#include <sys/resource.h>
#include <sys/wait.h>
#include <signal.h>

#include "globals.h"
#include "dll.h"
#include "dllapps.h"
#include "access.h"
#include "readconfig.h"
#include "fileio.h"
#include "netio.h"
#include "logfun.h"
#include "process.h"

/*
	check for abort condition
*/

INT	checkAbort(sopair_t *sockets, tftpR_t *req)
{
    if (isAborted) {
        if (!SilentDeny)
            sendToError(sockets, &req->reqPeer, PE_UNDEF);
        return 1;
    }
    return 0;
}

/*
	select the appropriate transfer type and process it
*/

INT processRequest (sopair_t *sockets, tftpR_t *req)
{
    INT rtc;

    if (getpid() != parentPid)
        deletePidChain (procListH);	/* get rid of parent entries */

    if ((rtc = validateAccess(req, 0))) {
        if (!SilentDeny)
            sendToError(sockets, &req->reqPeer, (rtc & (~PE_ERRMASK)));
        return 1;
    }

    if (checkAbort(sockets, req))
        return 1;

    if (connectPeer(req))
        return 1;

    if (checkAbort(sockets, req))
        return 1;

    if (req->reqType == RT_RRQ) {
        rtc = sendFile(req);
    }
    else if (req->reqType == RT_WRQ) {
        rtc = recvFile(req);
    }
    else
        rtc = 1;

    if (checkAbort(sockets, req))
        return 1;

    LOG_MSG3 (LOG_INFO, "Transferred %f MBytes %s %f MBytes disk storage",
              getCountDM(&req->xfrstat),
              (req->reqType == RT_WRQ) ? "to" : "from",
              getCountDM(&req->iostat));

    fclose (req->reqFd);
    close  (req->Rsock);    
    close  (req->Ssock);    

    return rtc;
}


/*
	main TFTPD server program
*/

/*	Note:	lines maked	 */
	/**/ 
/*	run in a child process	 */

INT main (INT argc, CHAR **argv)
{
    tftpR_t  req;			/* request data			*/
    sopair_t *sockets;			/* primary sockets		*/
    sopair_t ftsocks;			/* pr/socks if started by inetd */
    INT      rtc;                       /* getRequest return code	*/
    pid_t    pid;			/* child pid in process list	*/
    dlp_t    *pp;			/* process list entry		*/

/*
	Startup-Phase
	parse the command line input and set the globals consitently
	the configuration file is read & interpreted for the 1st time
	later, at the begin of each transfer, the configuration file
	is checked for any modigfications and re-parsed if necessary
	if only a configuration check is launched, we're finished here
*/

    evalCommandLine (argc, argv, 1);	/* get command line params	*/

    startLog();

    if (!CheckOnly)
        LOG_MSG3 (LOG_INFO, "Started at message level: General:%d, Syslog:%d",
                  MsgLevel, MaxSyslog);

    rtc = readConfigFile(ConfigFile);	/* parse tconfig file	*/

    if (CheckOnly) {
        if (rtc) {
            LOG_MSG1 (LOG_ERR, "Configuration check returns %d errors", rtc);
        }
        else {
            LOG_MSG4 (LOG_DEBUG, "Configuration checks o.k.");
        }
        exit (0);
    }
    else if (rtc)
        exit (1);

/*
	now we're sure there's some real work ahead
	we must decide, wich mode of operation should be entered
		- started by inetd,
		- standalone daemon or 
		- standalone / single-process (debug mode)
*/

        parentPid = getpid();

/*
	here on single process mode for debugging
	we run in an endless loop, processing all request serialized
*/

    if (spMode) {

        if ((sockets = netStart()) == NULL)	/* initialize network	*/
            exit (1);

        while (1) {

            if (readConfigFile(ConfigFile)) 	/* check for modific.	*/
                exit (1);

            rtc = getRequest(sockets, &req, MaxInputWait);

            if (!rtc) {
                LOG_MSG3 (LOG_INFO, "Exiting after %d minute%s idle",
                          MaxInputWait/60,
                          ((MaxInputWait/60) == 1) ? "" : "s");
                break;
            }
            else if (rtc < 0) {
                exit (1);
            }

            changeIdent();		/* change uid, gid & root	*/

            processRequest(sockets, &req);	/* do the final work	*/

            resetIdent();		/* switch back to parent u/gid	*/
        }
    }
/*
	this is the 'daemon' section
	each new transfer request causes the spawn of a new child process
	the child then changes UID/GID and the root if desired
	from this new (low privileged) incarnation of tftpd the access 
	validation takes place and, if successful, the transfer is done
*/
 
    else {

        if (isaSocket(stdin)) {		/* started by inetd ?		*/
            sockets = &ftsocks;
            sockets->recv = fileno(stdin);
            sockets->send = fileno(stdout);
            if (signalSetup(PARENT_SIGNALS))
                exit (1);

        }
        else {				/* no, we're standalone		*/

            if (becomeDaemon()) {
                processRundown();
                exit (1);
            }

            if ((sockets = netStart()) == NULL) { /* initialize network	*/
                processRundown();
                exit (1);
            }
        }

        parentPid = getpid();	/* update this to reflect our new status */

        while (1) {

            if (readConfigFile(ConfigFile))	/* check for modific.	*/
                break;

            rtc = getRequest(sockets, &req, MaxInputWait);

            if (!rtc) {
                INT children;
                if ((children = getProcNum())) {
                    LOG_MSG3 (LOG_INFO, 
                              "Still %d child%s busy after %d minute%s idle",
                              children,
                              (children > 1) ? "ren" : "", 
                              MaxInputWait/60,
                              ((MaxInputWait/60) == 1) ? "" : "s");
                    continue;
                }
                else {
                    LOG_MSG3 (LOG_INFO, "Exiting after %d minute%s idle",
                              MaxInputWait/60,
                              ((MaxInputWait/60) == 1) ? "" : "s");
                    break;
                }
            }
            else if (rtc < 0) {
                exit (1);
            }

	    if (!checkThreads(&req,&pp))/* max. # of threads running ?	*/
                continue;

            if ((pid = doFork(1))) {	/* parent 			*/
                modifyPLE(pp, pid, 0, 0, 0, 0);
                dumpProcList();
                continue;		/*	continues		*/
            }

      /**/  startLog();			/* assure our own log channel	*/
      /**/  signalReset(PARENT_SIGNALS);/* switch off special handling  */
      /**/  signalSetup(CHILD_SIGNALS);	/* switch on dump signal only	*/
      /**/  changeIdent();		/* change uid, gid & root	*/
      /**/
      /**/  rtc = processRequest(sockets, &req); /* do the real work	*/
      /**/
      /**/  exit (rtc);
        }
        processRundown();
    }
    return rtc;
}
